------------------ Global Variables -----------------
-- Table variables
CT_TrackedItemsTable = {}
CT_FilterSettings = {
    showConsumables = true,
    showContainers = false,
    showWeapons = false,
    showGems = true,
    showTradeskills = true,
    showEnhancements = true,
    showQuest = true,
    showGlyphs = false,
    showMisc = false,
}
CT_AutoRestock = {
    merchantRestock = false,
    bankRestock = false,
    guildBankRestock = false,
}


------------------ Local Variables ------------------

-- The last item the player clicked on
local selectedItem = nil

-- The last entered value for the input
local minAmtInputVal

-- The total number of slots in the interface
local numSlots = 500

-- Reference for color picker values
local initialR, initialG, initialB = 1, 1, 1



------------------ Helper Methods -------------------

-- Prints the itemTable to the console
local function displayCTitemTable()
    local safeColorR, safeColorG, safeColorB = 0.3, 1, 0.4
    local warningColorR, warningColorG, warningColorB = 1, 0.4, 0.4
    local dangerColorR, dangerColorG, dangerColorB = 1, 0.6, 0.2

    DEFAULT_CHAT_FRAME:AddMessage("=============== Tracked Consumables ===============", 1, 0.8, 0.3, 5)
    for _,item in pairs(CT_TrackedItemsTable) do
        local itemID, itemType = GetItemInfoInstant(item.link)
        
        if item.count == 0 then
            DEFAULT_CHAT_FRAME:AddMessage(item.link.." - "..itemType..", Amount: "..item.count.."/"..item.minAmt, warningColorR, warningColorG, warningColorB, 5)
        elseif item.count < item.minAmt then
        DEFAULT_CHAT_FRAME:AddMessage(item.link.." - "..itemType..", Amount: "..item.count.."/"..item.minAmt, dangerColorR, dangerColorG, dangerColorB, 5)
        else
            DEFAULT_CHAT_FRAME:AddMessage(item.link.." - "..itemType..", Amount: "..item.count.."/"..item.minAmt, safeColorR, safeColorG, safeColorB, 5)
        end
    end
    DEFAULT_CHAT_FRAME:AddMessage("=================================================", 1, 0.8, 0.3, 5)
end

-- Handles slash commands
local function ConsumableTracker_Commands(msg, editbox)
    if msg == "help" then
        print("/ct help - shows this list")
        print("/ct - opens the interface")
        print("/ct display - prints out a text representation of your tracked items")
    elseif msg == "display" then
        displayCTitemTable()
    else
        ShowUIPanel(ConsumableTracker)
        ConsumableTracker_Update()
    end
end

local function GetNextOpenSlotBag(nextBag) 
    if nextBag >= NUM_BAG_SLOTS then
        return -1
    end         

    if GetContainerNumFreeSlots(nextBag) > 0 then
        return nextBag
    else
        return GetNextOpenSlotBag(nextBag + 1) 
    end                        
end

-- Notifies the player of all items that fall below the minimum amount for that item
function ConsumableTracker_CheckUnderstock(item)
    if item.count == 0 then
        DEFAULT_CHAT_FRAME:AddMessage("You are completely out of "..item.link.."!", item.textR, item.textG, item.textB, 5)
        if item.soundEnabled then
            PlaySound(SOUNDKIT.TELL_MESSAGE, "SFX")
        end
        return true
    elseif item.count < (item.minAmt or 0) then
        DEFAULT_CHAT_FRAME:AddMessage("Your "..item.link.." has gone below your minimum amount of "..item.minAmt..". Remaining: ("..item.count..")", item.textR, item.textG, item.textB, 5)
        if item.soundEnabled then
            PlaySound(SOUNDKIT.TELL_MESSAGE, "SFX")
        end
        return true
    end
end

-- Is called by the XML when the addon is loaded
function ConsumableTracker_OnLoad(self)
    -- UIPanelWindows["ConsumableTracker"] = {
    --     area = "left",
    --     pushable = 1,
    --     whileDead = 1,
    -- }
    local frame = CreateFrame("Frame", "ConsumableTracker", UIParent)
    self.items = {}
    frame.name = "Consumable Tracker"
    frame.ok = function(self) ShowUIPanel(ConsumableTracker) end
    InterfaceOptions_AddCategory(frame)
    
    local numItemsInRow = 8
    local curRow = 1

    for idx=1, numSlots do
        local item = CreateFrame("Button", "ConsumableTracker_Item" .. idx, CTInvScrollFrame_Child, "ConsumableTracker_ItemTemplate")
        item:RegisterForClicks("RightButtonUp", "LeftButtonUp")

        self.items[idx] = item
        if idx == 1 then
            item:SetPoint("TOPLEFT", 4, -2)
        elseif idx % (numItemsInRow * curRow + 1) == 0 then
            item:SetPoint("TOPLEFT", self.items[idx-numItemsInRow], "BOTTOMLEFT", 0, -2)
            curRow = curRow + 1
        else
            item:SetPoint("TOPLEFT", self.items[idx-1], "TOPRIGHT", 2, 0)
        end
    end

    for item, entry in pairs(CT_TrackedItemsTable) do
        -- Reset the count for the tracked items so the count is accurate after looping through bags
        CT_TrackedItemsTable[item].count = 0;
    end

    ConsumableTracker_CreateFiltersDropdownMenu()
    ConsumableTracker_CreateContextsDropdownMenu()

    local frameTitle=CreateFrame("Frame", "CT_Title", frame)
	frameTitle:SetPoint("TOPLEFT", 10, -10)
	frameTitle:SetWidth(300)
	frameTitle:SetHeight(100)
	frameTitle:Show()

    local frameTitleString = frameTitle:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
    frameTitleString:SetPoint("TOPLEFT", 0, 0)
    frameTitleString:SetText('|cff3aff69Consumable Tracker|r')

    frame.displayWindowBtn = CreateFrame ("Button", nil, frame, "OptionsButtonTemplate")
	frame.displayWindowBtn:SetText ("Show Interface")
	frame.displayWindowBtn:SetPoint ("TOPLEFT", frame, "TOPLEFT", 20, -60)
	frame.displayWindowBtn:SetWidth (150)
	frame.displayWindowBtn:SetScript ("OnClick", function (self) ShowUIPanel(ConsumableTracker) end)

    frame.printTableBtn = CreateFrame ("Button", nil, frame, "OptionsButtonTemplate")
	frame.printTableBtn:SetText ("Print Tracked Items List")
	frame.printTableBtn:SetPoint ("TOPLEFT", frame, "TOPLEFT", 20, -100)
	frame.printTableBtn:SetWidth (150)
	frame.printTableBtn:SetScript ("OnClick", function (self) displayCTitemTable() end)

    frame.restockFromMerchantsChk = CreateFrame("CheckButton", "ConsumableTracker_MerchantRestockChk", frame, "InterfaceOptionsCheckButtonTemplate")
    ConsumableTracker_MerchantRestockChkText:SetText ("Auto Restock from Vendors")
    frame.restockFromMerchantsChk:SetPoint("TOPLEFT", frame, "TOPLEFT", 20, -140)
    frame.restockFromMerchantsChk.tooltip = "While enabled, you will purchase understocked tracked items from vendors that sell the item. (Buys up to the item's minimum.)"
    frame.restockFromMerchantsChk:SetScript("OnClick", function(self) CT_AutoRestock.merchantRestock = not CT_AutoRestock.merchantRestock; end)
    frame.restockFromMerchantsChk:SetChecked(CT_AutoRestock.merchantRestock)

    frame.restockFromBankChk = CreateFrame("CheckButton", "ConsumableTracker_BankRestockChk", frame, "InterfaceOptionsCheckButtonTemplate")
    ConsumableTracker_BankRestockChkText:SetText ("Auto Restock from Bank")
    frame.restockFromBankChk:SetPoint("TOPLEFT", frame, "TOPLEFT", 20, -180)
    frame.restockFromBankChk.tooltip = "While enabled, you will withdraw understocked tracked items from your bank. (Withdraws up to the item's minimum.)"
    frame.restockFromBankChk:SetScript("OnClick", function(self) CT_AutoRestock.bankRestock = not CT_AutoRestock.bankRestock; end)
    frame.restockFromBankChk:SetChecked(CT_AutoRestock.bankRestock)

    -- frame.restockFromGuildBankChk = CreateFrame("CheckButton", "ConsumableTracker_GuildBankRestockChk", frame, "InterfaceOptionsCheckButtonTemplate")
    -- ConsumableTracker_GuildBankRestockChkText:SetText ("Auto Restock from Guild Bank")
    -- frame.restockFromGuildBankChk:SetPoint("TOPLEFT", frame, "TOPLEFT", 20, -220)
    -- frame.restockFromGuildBankChk.tooltip = "While enabled, you will withdraw understocked tracked items from your guild bank. (Withdraws up to the item's minimum.)"
    -- frame.restockFromGuildBankChk:SetScript("OnClick", function(self) CT_AutoRestock.guildBankRestock = not CT_AutoRestock.guildBankRestock; end)
    -- frame.restockFromGuildBankChk:SetChecked(CT_AutoRestock.guildBankRestock)

    self.bagCounts = {}
    self:RegisterEvent("ADDON_LOADED")
end

-- Called when the Addon is first loaded, and also when a related event is fired
function ConsumableTracker_OnEvent(self, event, ...)
    if event == "ADDON_LOADED" and ... == "ConsumableTracker" then
        for bag = 0, NUM_BAG_SLOTS do
            -- Use the optional flag to skip updating times
            ConsumableTracker_ScanBag(bag, true)
        end
        
        ConsumableTracker_Update()
        
        local size = 0
        for link, entry in pairs(CT_TrackedItemsTable) do
            size = size + 1
            -- Reset timer variables to prevent issues with GetTime() caused by computer restarts
            CT_TrackedItemsTable[link].curTime = 0
            CT_TrackedItemsTable[link].lastUpdate = 0
            CT_TrackedItemsTable[link].delta = 0

            -- Trigger a notification for all understock items on load, even if not repeating timers
            CT_TrackedItemsTable[link].shouldFire = true
        end
        -- If the player is not tracking any items, show the bag items by default
        if size == 0 then
            ConsumableTracker_DisplayTrackedOnly:SetChecked(false)
        end

        ConsumableTracker_InitializeSlider(ConsumableTracker_SettingsFrame_TimerSlider)

        local tickerHandler = C_Timer.NewTicker(1, ConsumableTracker_Tick)

        ConsumableTracker_MerchantRestockChk:SetChecked(CT_AutoRestock.merchantRestock)
        ConsumableTracker_BankRestockChk:SetChecked(CT_AutoRestock.bankRestock)
        --ConsumableTracker_GuildBankRestockChk:SetChecked(CT_AutoRestock.guildBankRestock)

        self:UnregisterEvent("ADDON_LOADED")
        self:RegisterEvent("BAG_UPDATE")
        self:RegisterEvent("BAG_UPDATE_COOLDOWN")
        self:RegisterEvent("ITEM_PUSH")
        self:RegisterEvent("UNIT_INVENTORY_CHANGED")
        self:RegisterEvent("ITEM_LOCK_CHANGED")
        self:RegisterEvent("PLAYER_LOGOUT")
        self:RegisterEvent("PLAYER_ENTERING_WORLD")
        self:RegisterEvent("MERCHANT_SHOW")
        self:RegisterEvent("BANKFRAME_OPENED")
        self:RegisterEvent("GUILDBANKFRAME_OPENED")
    elseif event == "BAG_UPDATE" or event == "BAG_UPDATE_COOLDOWN" then
        local bag = ...
        if bag and bag >= 0 then
            ConsumableTracker_ScanBag(bag)
            ConsumableTracker_Update()
        end
    elseif event == "UNIT_INVENTORY_CHANGED" or event == "ITEM_PUSH" or event == "PLAYER_ENTERING_WORLD" or event == "ITEM_LOCK_CHANGED" then
        ConsumableTracker_Update()
    elseif event == "MERCHANT_SHOW" and CT_AutoRestock.merchantRestock then
        ConsumableTracker_RestockItemsFromMerchant()
    elseif event == "BANKFRAME_OPENED" and CT_AutoRestock.bankRestock then
        ConsumableTracker_RestockItemsFromBank()
    elseif event == "GUILDBANKFRAME_OPENED" and CT_AutoRestock.guildBankRestock then
        ConsumableTracker_RestockItemsFromGuildBank()
    end
end

-- Handles scanning bags in order to retrieve data for the player's items. Uses this data to populate the interface.
function ConsumableTracker_Update()
    local size = 0
    for link, entry in pairs(CT_TrackedItemsTable) do
        size = size + 1
        ConsumableTracker_DisplayTrackedOnly:Show()
        break
    end
    -- If the player is not tracking any items, do not allow them to only show tracked items
    if size == 0 then
        ConsumableTracker_DisplayTrackedOnly:SetChecked(false)
        ConsumableTracker_DisplayTrackedOnly:Hide()
    end
    


    local bagItems = {}
    if ConsumableTracker_DisplayTrackedOnly:GetChecked() == false then
        for bag = 0, NUM_BAG_SLOTS do
            for slot = 0, GetContainerNumSlots(bag) do
                local texture, count, locked, quality, readable, lootable, link = GetContainerItemInfo(bag, slot)
                if texture then
                    local shown = false
                    local itemID, itemType = GetItemInfoInstant(link)
                    if (itemType == "Consumable" and CT_FilterSettings.showConsumables) or 
                    (itemType == "Container" and CT_FilterSettings.showContainers) or
                    (itemType == "Weapon" and CT_FilterSettings.showWeapons) or
                    (itemType == "Gem" and CT_FilterSettings.showGems) or
                    (itemType == "Armor" and CT_FilterSettings.showArmor) or
                    (itemType == "Tradeskill" and CT_FilterSettings.showTradeskills) or
                    (itemType == "Item Enhancement" and CT_FilterSettings.showEnhancements) or
                    (itemType == "Quest" and CT_FilterSettings.showQuest) or
                    (itemType == "Glyph" and CT_FilterSettings.showGlyphs) or
                    (itemType == "Miscellaneous" and CT_FilterSettings.showMisc)
                     then
                        shown = true
                    end

                    if shown then
                        local itemNum = tonumber(link:match("|Hitem:(%d+):"))
                        if not bagItems[itemNum] then
                            bagItems[itemNum] = {
                                texture = texture,
                                count = count,
                                quality = quality, 
                                link = link,
                                num = itemNum,
                                tracked = false,
                                minAmt = 0,
                                timerInterval = 10,
                                curTime = 0,
                                lastUpdate = 0,
                                delta = 0,
                                soundEnabled = false,
                                notifSound = nil,
                                textR = 1,
                                textB = 1,
                                textG = 1,
                                shouldFire = true,
                                repeating = true,
                                showInRaid = true,
                                showInGroup = true,
                                showUngrouped = true,
                            }
                        else
                            bagItems[itemNum].count = bagItems[itemNum].count + count
                        end
                    end
                end
            end
        end
    else
        bagItems = {}
        local itemCounts = {}
        for bag = 0, NUM_BAG_SLOTS do
            for slot = 0, GetContainerNumSlots(bag) do
                local texture, count, locked, quality, readable, lootable, link = GetContainerItemInfo(bag, slot)
                if texture then
                    local itemId = tonumber(link:match("|Hitem:(%d+):"))
                    if itemCounts[itemId] then
                        itemCounts[itemId] = itemCounts[itemId] + count
                    else
                        itemCounts[itemId] = count
                    end
                end
            end
        end

        for link, entry in pairs(CT_TrackedItemsTable) do
            if entry.count ~= (itemCounts[entry.num]) then
                CT_TrackedItemsTable[link].count = (itemCounts[entry.num] or 0)
            end
        end 
    end

    local displayItems = {}
    local numOfItems = 0
    for link, entry in pairs(CT_TrackedItemsTable) do
        numOfItems = numOfItems + 1
        local inputText = ConsumableTracker_SearchField_Search:GetText()
        if #inputText > 0 then
            if(string.find(string.lower(entry.link), string.lower(inputText), 1, true)) then
                table.insert(displayItems, entry)
                bagItems[entry.num] = nil
            end
        else
            table.insert(displayItems, entry)
            bagItems[entry.num] = nil
        end
        
    end

    for link, entry in pairs(bagItems) do
        numOfItems = numOfItems + 1
        local inputText = ConsumableTracker_SearchField_Search:GetText()
        if #inputText > 0 then
            if(string.find(string.lower(entry.link), string.lower(inputText), 1, true)) then
                table.insert(displayItems, entry)
            end
        else
            table.insert(displayItems, entry)
        end
    end

    CTInvScrollFrame_VSlider:SetMinMaxValues(0, numOfItems)

    for idx = 1, numSlots do
        local button = ConsumableTracker.items[idx]
        local entry = displayItems[idx]

        if entry then
            button.item = entry
            if button.item.tracked then
                button.glow:Show()
            else
                button.glow:Hide()
            end
            button:SetAttribute("item2", entry.name)
            button.link = entry.link
            button.icon:SetTexture(entry.texture)
            -- If the item count is greater than 1, display it
            if entry.count > 1 then
                button.count:SetText(entry.count)
                button.count:Show()
            else
                -- Always show the exact count of tracked items, but hide
                -- the count if the item is not tracked (default WoW behavior)
                if not button.item.tracked then
                    button.count:Hide()
                else
                    button.count:SetText(entry.count)
                    button.count:Show()
                end
            end
            button:Show()
        else
            button.link = nil
            button:Hide()
        end
    end
end

-- Progresses every active timer by 1 second (called from a repeating Ticker)
function ConsumableTracker_Tick()
    for link, entry in pairs(CT_TrackedItemsTable) do
        if CT_TrackedItemsTable[link].curTime then
            if(CT_TrackedItemsTable[link].curTime < (CT_TrackedItemsTable[link].timerInterval)) then
                local now = GetTime()
                CT_TrackedItemsTable[link].delta = now - CT_TrackedItemsTable[link].lastUpdate
                CT_TrackedItemsTable[link].lastUpdate = now
                CT_TrackedItemsTable[link].curTime = CT_TrackedItemsTable[link].curTime + CT_TrackedItemsTable[link].delta
            else
                CT_TrackedItemsTable[link].curTime = 0
                CT_TrackedItemsTable[link].delta = 0

                if CT_TrackedItemsTable[link].shouldFire then
                    shouldShow = CT_TrackedItemsTable[link].showInRaid and IsInRaid()
                    if not shouldShow then
                        shouldShow = CT_TrackedItemsTable[link].showInGroup and IsInGroup()
                        if not shouldShow then
                            shouldShow = CT_TrackedItemsTable[link].showUngrouped
                        end 
                    end

                    if shouldShow then
                        if ConsumableTracker_CheckUnderstock(CT_TrackedItemsTable[link]) then
                            if CT_TrackedItemsTable[link].repeating == false then
                                CT_TrackedItemsTable[link].shouldFire = false
                            end
                        end 
                    end
                end
            end
        else
            CT_TrackedItemsTable[link].curTime = 0
            CT_TrackedItemsTable[link].lastUpdate = 0
            CT_TrackedItemsTable[link].delta = 0
        end
    end
end

-- Called from OnEvent
function ConsumableTracker_ScanBag(bag, initial)
    if not ConsumableTracker.bagCounts[bag] then
        ConsumableTracker.bagCounts[bag] = 0
    end

    -- Loop through all of the slots and store the total count of each item
    local itemCounts = {}
    for slot = 0, GetContainerNumSlots(bag) do
        local texture, count, locked, quality, readable, lootable, link = GetContainerItemInfo(bag, slot)
        if texture then
            local itemId = tonumber(link:match("|Hitem:(%d+):"))
            if itemCounts[itemId] then
                itemCounts[itemId] = itemCounts[itemId] + count
            else
                itemCounts[itemId] = count
            end
        end
    end

    -- Check to see if this is the first scan, as we would have nothing to compare to
    if initial then
        for itemId, count in pairs(itemCounts) do
            -- Store the time the item was added (either the previous value, if possible, or the current time)
            --ConsumableTracker_ItemTimes[itemId] = ConsumableTracker_ItemTimes[itemId] or time()
        end
    else
        for itemId, count in pairs(itemCounts) do
            local oldCount = ConsumableTracker.bagCounts[bag][itemId] or 0
            if count > oldCount then
                --ConsumableTracker_ItemTimes[itemId] = time()
            end
        end
    end

    ConsumableTracker.bagCounts[bag] = itemCounts
end

-- Sets the warning notification threshold value for the selected item.
function ConsumableTracker_SetMinNum()
    --local minSlider = ConsumableTracker_SettingsFrame_MinNumSlider
    
    local val = tonumber(ConsumableTracker_SettingsFrame.input:GetText())

    if val then
        val = math.floor(val)
        --minSlider:SetValue(val)
        if val > 99999 then
            val = 99999
        elseif val < 0 then
            val = 0
        end

        if selectedItem then
            selectedItem.minAmt = val
            if(CT_TrackedItemsTable[selectedItem.num]) then
                CT_TrackedItemsTable[selectedItem.num].minAmt = val
            end
        end

        ConsumableTracker_SettingsFrame.input:SetText(("%d"):format(val, 0))
    else
        ConsumableTracker_SettingsFrame.input:SetText(minAmtInputVal)
    end

    ConsumableTracker_SettingsFrame.input:ClearFocus()
end

function ConsumableTracker_RestockItemsFromMerchant()
    for i = 1, GetMerchantNumItems() do
        local mItemName = GetMerchantItemInfo(i)
        for link, entry in pairs(CT_TrackedItemsTable) do
            local tItemName = GetItemInfo(CT_TrackedItemsTable[link].link)
            if(tItemName == mItemName) then
                local neededCnt = CT_TrackedItemsTable[link].minAmt - CT_TrackedItemsTable[link].count
                if(neededCnt > 0) then
                    DEFAULT_CHAT_FRAME:AddMessage("CT: Attempted to purchase " .. neededCnt .. " " .. CT_TrackedItemsTable[link].link..".", CT_TrackedItemsTable[link].textR, CT_TrackedItemsTable[link].textG, CT_TrackedItemsTable[link].textB, 5)
                    BuyMerchantItem(i, neededCnt)
                end
            end
        end
    end
end

function ConsumableTracker_RestockItemsFromBank()
    -- Loop through each tracked item and set how below the min that item is
    for link, entry in pairs(CT_TrackedItemsTable) do
        local trackedItem = CT_TrackedItemsTable[link]
        local neededCnt = trackedItem.minAmt - trackedItem.count
        -- Loop through all of the bank bags, checking each bank container for the item
        for bag = BANK_CONTAINER, NUM_BAG_SLOTS+NUM_BANKBAGSLOTS do
            if bag == BANK_CONTAINER or bag > NUM_BAG_SLOTS then
                for slot = 1, GetContainerNumSlots(bag) do
                    local texture, count, locked, quality, readable, lootable, mItemlink = GetContainerItemInfo(bag, slot)
                    if(trackedItem.link == mItemlink) then
                        -- If the stack contains less than we need, take what there is
                        -- Otherwise, split the stack and place it in the next open slot in the inventory.
                        if(neededCnt > 0) then
                            if count <= neededCnt then
                                DEFAULT_CHAT_FRAME:AddMessage("CT: Attempted to withdraw " .. neededCnt .. " " .. CT_TrackedItemsTable[link].link..".", CT_TrackedItemsTable[link].textR, CT_TrackedItemsTable[link].textG, CT_TrackedItemsTable[link].textB, 5)
                                UseContainerItem(bag, slot)
                                neededCnt = neededCnt - count
                                trackedItem.count = trackedItem.count + count
                            else
                                SplitContainerItem(bag, slot, neededCnt)
                                local nextOpenBag = GetNextOpenSlotBag(BACKPACK_CONTAINER)
                                if nextOpenBag > -1 then
                                    local openSlots = GetContainerFreeSlots(nextOpenBag)
                                    DEFAULT_CHAT_FRAME:AddMessage("CT: Attempted to withdraw " .. neededCnt .. " " .. CT_TrackedItemsTable[link].link..".", CT_TrackedItemsTable[link].textR, CT_TrackedItemsTable[link].textG, CT_TrackedItemsTable[link].textB, 5)
                                    PickupContainerItem(nextOpenBag, openSlots[1])
                                    neededCnt = 0
                                    trackedItem.count = trackedItem.count + neededCnt
                                    nextOpenBag = GetNextOpenSlotBag(BACKPACK_CONTAINER)
                                end
                                ClearCursor()
                            end
                        end
                    end
                end
            end
        end
    end
end

function ConsumableTracker_RestockItemsFromGuildBank()
    -- local numTabs = GetNumGuildBankTabs()
    -- local currentTab = GetCurrentGuildBankTab()
    -- local MAX_GUILDBANK_SLOTS_PER_TAB = 98

    -- -- Loop through each tracked item and set how below the min that item is
    -- for link, entry in pairs(CT_TrackedItemsTable) do
    --     local trackedItem = CT_TrackedItemsTable[link]
    --     local neededCnt = trackedItem.minAmt - trackedItem.count
    --     -- Loop through all of the bank bags, checking each bank container for the item
    --     for tab = 1, MAX_GUILDBANK_TABS do
    --         for slot = 1, MAX_GUILDBANK_SLOTS_PER_TAB do
    --             local texture, count, locked, isFiltered, quality = GetGuildBankItemInfo(tab,slot)
    --             local mItemlink
    --             if texture then
    --                 mItemlink = GetGuildBankItemLink(tab,slot)
    --             end
    --             if(trackedItem.link == mItemlink) then
    --                 -- If the stack contains less than we need, take what there is
    --                 if(neededCnt > 0) then
    --                     local amt
    --                     if count < neededCnt then
    --                         amt = count
    --                     else
    --                         amt = neededCnt
    --                     end
    --                     SplitGuildBankItem(tab, slot, amt)
    --                     DEFAULT_CHAT_FRAME:AddMessage("CT: Attempted to withdraw " .. amt .. " " .. trackedItem.link..".", trackedItem.textR, trackedItem.textG, trackedItem.textB, 5)
    --                     local infoType, itemID, itemLink = GetCursorInfo()
    --                     if infoType == "item" then
    --                         -- Find the bag the item exists in and place it there or in the first open spot if count is 0
    --                         for iBag = BACKPACK_CONTAINER, NUM_BAG_SLOTS do
    --                             if neededCnt > 0 then 
    --                                 for iSlot = 1, GetContainerNumSlots(iBag) do
    --                                     local _, stackCount, _, _, _, _, iItemlink = GetContainerItemInfo(iBag, iSlot)
    --                                     if (iItemlink == trackedItem.link) or (iItemlink == nil and trackedItem.count == 0) then
    --                                         PickupContainerItem(iBag, iSlot)
    --                                         neededCnt = neededCnt - amt
    --                                         trackedItem.count = trackedItem.count + amt
    --                                     end
    --                                 end
    --                             end
    --                         end
    --                     end
    --                     ClearCursor()
    --                 end
    --             end
    --         end
    --     end
    -- end
end

------------------ UI Methods -----------------------
-- Called when one of the interface item slots is clicked
function ConsumableTracker_Button_OnClick(self, button, down)
    selectedItem = self.item
    ConsumableTracker_SettingsFrame:Show()
    ConsumableTracker_SettingsFrame_SelectedItem_Texture:SetTexture(selectedItem.texture)
    ConsumableTracker_SettingsFrame_SelectedItem_Text:SetText(selectedItem.link)
    ConsumableTracker_SettingsFrame_IsTracked:SetChecked(selectedItem.tracked)
    ConsumableTracker_SettingsFrame.input:SetText(selectedItem.minAmt or 0)
    ConsumableTracker_SettingsFrame_TimerSlider:SetValue(selectedItem.timerInterval / 60 or 10)
    ConsumableTracker_SettingsFrame_EnableSound:SetChecked(selectedItem.soundEnabled)
    ConsumableTracker_SettingsFrame_ExampleText_Text:SetTextColor(selectedItem.textR, selectedItem.textG, selectedItem.textB)
    ConsumableTracker_SettingsFrame_Repeating:SetChecked(selectedItem.repeating)
    CloseDropDownMenus()
end

-- Shows the color picker
function ConsumableTracker_SelectColorButton_OnClick(self, button, down)
    ShowColorPicker(selectedItem.textR, selectedItem.textG, selectedItem.textB, nil, ConsumableTracker_ColorPicker_Callback)
end

-- Overrides the WoW API function that displays the color picker.
function ShowColorPicker(r, g, b, a, changedCallback)
    ColorPickerFrame.hasOpacity, ColorPickerFrame.opacity = (a ~= nil), a
    ColorPickerFrame.previousValues = {r,g,b,a}
    ColorPickerFrame.func, ColorPickerFrame.opacityFunc, ColorPickerFrame.cancelFunc = changedCallback, changedCallback, changedCallback
    ColorPickerFrame:SetColorRGB(r,g,b)
    ColorPickerFrame:Hide() -- Need to run the OnShow handler.
    ColorPickerFrame:Show()
end

-- Manages the logic of the color picker
function ConsumableTracker_ColorPicker_Callback(restore)
    local newR, newG, newB, newA
    if restore then
     -- The user bailed, we extract the old color from the table created by ShowColorPicker.
     newR, newG, newB = unpack(restore)
    else
     -- Something changed
     newA, newR, newG, newB = OpacitySliderFrame:GetValue(), ColorPickerFrame:GetColorRGB()
    end
    
    -- Update our internal storage.
    initialR, initialG, initialB = newR, newG, newB
    -- And update any UI elements that use this color...
    selectedItem.textR = initialR
    selectedItem.textG = initialG
    selectedItem.textB = initialB
    if CT_TrackedItemsTable[selectedItem.num] then
        CT_TrackedItemsTable[selectedItem.num].textR = selectedItem.textR
        CT_TrackedItemsTable[selectedItem.num].textG = selectedItem.textG
        CT_TrackedItemsTable[selectedItem.num].textB = selectedItem.textB
    end
    ConsumableTracker_SettingsFrame_ExampleText_Text:SetTextColor(selectedItem.textR, selectedItem.textG, selectedItem.textB)
end

-- Called when the player clicks on the "Is Tracked" setting checkbox
function ConsumableTracker_IsTracked_OnClick()
    if selectedItem then
        selectedItem.tracked = ConsumableTracker_SettingsFrame_IsTracked:GetChecked()

        -- If the user checks "Is Tracked" add it to the table, else, remove it from the table
        if selectedItem.tracked then
            if not CT_TrackedItemsTable[selectedItem.num] then
                CT_TrackedItemsTable[selectedItem.num] = selectedItem
            end
        else
            if CT_TrackedItemsTable[selectedItem.num] then
                CT_TrackedItemsTable[selectedItem.num] = nil
            end
        end

        -- Hacky way of updating glow on buttons. Should be done for just that button.
        ConsumableTracker_Update()
    end
end

-- Called when the player clicks on the "Display Tracked Items Only" setting checkbox
function ConsumableTracker_DisplayTrackedOnly_OnClick(self, button)
    ConsumableTracker_Update()
end

-- Scrolls the inventory scrollframe using the mouse wheel
function ConsumableTracker_OnMouseWheel(self, delta)
    local current = CTInvScrollFrame_VSlider:GetValue()
    local stepVal = CTInvScrollFrame_VSlider:GetHeight() / 10

    local sliderMin, sliderMax = CTInvScrollFrame_VSlider:GetMinMaxValues()
    if (delta < 0) and (current < sliderMax) then
        CTInvScrollFrame_VSlider:SetValue(current + stepVal)
    elseif (delta > sliderMin) and (current > 1) then
        CTInvScrollFrame_VSlider:SetValue(current - stepVal)
    end
end

-- Used to set the proper values for the notification interval slider
function ConsumableTracker_InitializeSlider(slider)
    slider:ApplyBackdrop()
    slider:SetObeyStepOnDrag(true)
    val = 10
    slider:SetValue(val)
    slider.timerVal:SetText(slider:GetValue())
    sliderMin, sliderMax = slider:GetMinMaxValues()
    _G[slider:GetName().."Text"]:SetText("Warning Interval (mins):")
    _G[slider:GetName().."Low"]:SetText(sliderMin)
    _G[slider:GetName().."High"]:SetText(sliderMax)
end

-- Called when the player adjusts the value of the notification interval slider
function ConsumableTracker_UpdateSlider(slider)
    if selectedItem then
        slider.timerVal:SetText(slider:GetValue())
        selectedItem.timerInterval = slider:GetValue() * 60 -- * 60 for minutes
        if CT_TrackedItemsTable[selectedItem.num] then
            CT_TrackedItemsTable[selectedItem.num].timerInterval = selectedItem.timerInterval
        end
    end
end

-- Called when the player clicks on the input in the item settings tab
function ConsumableTracker_Input_OnFocusGained()
    minAmtInputVal = tonumber(ConsumableTracker_SettingsFrame.input:GetText())
end

-- Sets the minimum value set in the edit box and clears the focus
function ConsumableTracker_ApplyButton_OnClick(self, button, down)
    ConsumableTracker_SettingsFrame.input:ClearFocus()
    ConsumableTracker_SetMinNum()
end

-- Toggles the setting that handles the set sound for the notification
function ConsumableTracker_EnableSound_OnClick()
    if selectedItem then
        selectedItem.soundEnabled = ConsumableTracker_SettingsFrame_EnableSound:GetChecked()
        if CT_TrackedItemsTable[selectedItem.num] then
            CT_TrackedItemsTable[selectedItem.num].soundEnabled = selectedItem.soundEnabled
        end
    end
end

function ConsumableTracker_EnableRepeating_OnClick()
    if selectedItem then
        selectedItem.repeating = ConsumableTracker_SettingsFrame_Repeating:GetChecked()

        if selectedItem.repeating == true then
            selectedItem.shouldFire = true
        end

        if CT_TrackedItemsTable[selectedItem.num] then
            CT_TrackedItemsTable[selectedItem.num].repeating = selectedItem.repeating
            CT_TrackedItemsTable[selectedItem.num].shouldFire = selectedItem.shouldFire
        end
    end
end

function ConsumableTracker_ShowFilters_OnClick(self, button, down)
    ToggleDropDownMenu(1, nil, ConsumableTracker_FilterMenu, self:GetName(), 0, 0)
end

function ConsumableTracker_ShowContexts_OnClick(self, button, down)
    ToggleDropDownMenu(1, nil, ConsumableTracker_ContextsMenu, self:GetName(), 0, 0)
end

function ConsumableTracker_CreateFiltersDropdownMenu()
    local info = {}
    local ConsumableTracker_FilterMenu = CreateFrame("Frame", "ConsumableTracker_FilterMenu")
    ConsumableTracker_FilterMenu.displayMode = "MENU"
    ConsumableTracker_FilterMenu.initialize = function(self, level) 
        if not level then return end
        wipe(info)
        if level == 1 then
            info.isTitle = 1
            info.text = "Filter Item Types"
            info.notCheckable = 1;
            UIDropDownMenu_AddButton(info, level)

            info.disabled = false
            info.isTitle = false
            info.notCheckable = false
            info.keepShownOnClick = true

            info.text = "Consumable"
            info.func = function() CT_FilterSettings.showConsumables = not CT_FilterSettings.showConsumables
                ConsumableTracker_Update()
            end
            info.checked = CT_FilterSettings.showConsumables
            UIDropDownMenu_AddButton(info, level)

            info.text = "Container"
            info.checked = CT_FilterSettings.showContainers
            info.func = function() CT_FilterSettings.showContainers = not CT_FilterSettings.showContainers
                ConsumableTracker_Update()
            end
            UIDropDownMenu_AddButton(info, level)

            info.text = "Weapon"
            info.checked = CT_FilterSettings.showWeapons
            info.func = function() CT_FilterSettings.showWeapons = not CT_FilterSettings.showWeapons
                ConsumableTracker_Update()
            end
            UIDropDownMenu_AddButton(info, level)

            info.text = "Gem"
            info.checked = CT_FilterSettings.showGems
            info.func = function() CT_FilterSettings.showGems = not CT_FilterSettings.showGems
                ConsumableTracker_Update()
            end
            UIDropDownMenu_AddButton(info, level)

            info.text = "Armor"
            info.checked = CT_FilterSettings.showArmor
            info.func = function() CT_FilterSettings.showArmor = not CT_FilterSettings.showArmor
                ConsumableTracker_Update()
            end
            UIDropDownMenu_AddButton(info, level)

            info.text = "Tradeskill"
            info.checked = CT_FilterSettings.showTradeskills
            info.func = function() CT_FilterSettings.showTradeskills = not CT_FilterSettings.showTradeskills
                ConsumableTracker_Update()
            end
            UIDropDownMenu_AddButton(info, level)

            info.text = "Quest"
            info.checked = CT_FilterSettings.showQuest
            info.func = function() CT_FilterSettings.showQuest = not CT_FilterSettings.showQuest
                ConsumableTracker_Update()
            end
            UIDropDownMenu_AddButton(info, level)

            info.text = "Item Enhancement"
            info.checked = CT_FilterSettings.showEnhancements
            info.func = function() CT_FilterSettings.showEnhancements = not CT_FilterSettings.showEnhancements
                ConsumableTracker_Update()
            end
            UIDropDownMenu_AddButton(info, level)

            info.text = "Glyph"
            info.checked = CT_FilterSettings.showGlyphs
            info.func = function() CT_FilterSettings.showGlyphs = not CT_FilterSettings.showGlyphs 
                ConsumableTracker_Update()
            end
            UIDropDownMenu_AddButton(info, level)

            info.text = "Miscellaneous"
            info.checked = CT_FilterSettings.showMisc
            info.func = function() CT_FilterSettings.showMisc = not CT_FilterSettings.showMisc 
                ConsumableTracker_Update()
            end
            UIDropDownMenu_AddButton(info, level)
        end
    end
end

function ConsumableTracker_CreateContextsDropdownMenu()
    local info = {}
    local ConsumableTracker_ContextsMenu = CreateFrame("Frame", "ConsumableTracker_ContextsMenu")
    ConsumableTracker_ContextsMenu.displayMode = "MENU"
    ConsumableTracker_ContextsMenu.initialize = function(self, level) 
        if not level then return end
        wipe(info)
        if level == 1 then
            info.isTitle = 1
            info.text = "Show notifications while:"
            info.notCheckable = 1;
            UIDropDownMenu_AddButton(info, level)

            info.disabled = false
            info.isTitle = false
            info.notCheckable = false
            info.keepShownOnClick = true

            info.text = "In Raid"
            info.func = function() selectedItem.showInRaid = not selectedItem.showInRaid
            end
            info.checked = selectedItem.showInRaid
            info.tooltipTitle = "Show in Raids"
            info.tooltipText = "While checked, you will receive notifications for this item while in a raid group."
            info.tooltipOnButton = true
            UIDropDownMenu_AddButton(info, level)

            info.text = "In Party"
            info.checked = selectedItem.showInGroup
            info.tooltipTitle = "Show in Parties"
            info.tooltipText = "While checked, you will receive notifications for this item while in a party."
                info.tooltipOnButton = true
            info.func = function() selectedItem.showInGroup = not selectedItem.showInGroup
            end
            UIDropDownMenu_AddButton(info, level)

            info.text = "Solo"
            info.checked = selectedItem.showUngrouped
            info.tooltipTitle = "Show while Solo"
            info.tooltipText = "While checked, you will receive notifications for this item while not in a group."
            info.tooltipOnButton = true
            info.func = function() selectedItem.showUngrouped = not selectedItem.showUngrouped
            end
            UIDropDownMenu_AddButton(info, level)
        end
    end
end
------------------ Tooltip Methods -----------------------

-- Called when the mouse enters an item slot in the interface
function ConsumableTracker_Button_OnEnter(self, motion)
    if self.link then
        GameTooltip:SetOwner(self, "ANCHOR_TOPRIGHT")
        GameTooltip:SetHyperlink(self.link)
        GameTooltip:Show()
    end
    
end

-- Called when the mouse leaves an item slot in the interface
function ConsumableTracker_Button_OnLeave(self, motion)
    GameTooltip:Hide()
end

function ConsumableTracker_DisplayTrackedOnly_OnEnter(self, motion)
    GameTooltip:SetOwner(self, "ANCHOR_TOPRIGHT")
    GameTooltip:SetText("Display Tracked Items Only")
    GameTooltip:AddLine("While checked, only tracked items will be displayed in the interface.",1,1,1,true)
    GameTooltip:Show()
end

function ConsumableTracker_IsTracked_OnEnter(self, motion)
    GameTooltip:SetOwner(self, "ANCHOR_TOPRIGHT")
    GameTooltip:SetText("Is Tracked")
    GameTooltip:AddLine("While checked, this item will be tracked and you will receive notifications when its count in your inventory goes below your minimum or reaches zero.",1,1,1,true)
    GameTooltip:Show()
end

function ConsumableTracker_Input_OnEnter(self, motion)
    GameTooltip:SetOwner(self, "ANCHOR_TOPRIGHT")
    GameTooltip:SetText("Warn If Below")
    GameTooltip:AddLine("Set the minimum amount of this item before you are notified. \nIf the count of the item in your bags goes below this value, you will receive a notification. \nYou will always receive a notification if the count reaches zero for a tracked item.",1,1,1,true)
    GameTooltip:Show()
end

function ConsumableTracker_TimerSlider_OnEnter(self, motion)
    GameTooltip:SetOwner(self, "ANCHOR_TOPRIGHT")
    GameTooltip:SetText("Warning Interval")
    GameTooltip:AddLine("Set the number of minutes between notifications for this item.",1,1,1,true)
    GameTooltip:Show()
end

function ConsumableTracker_EnableSound_OnEnter(self, motion)
    GameTooltip:SetOwner(self, "ANCHOR_TOPRIGHT")
    GameTooltip:SetText("Enable Sound")
    GameTooltip:AddLine("If checked, a sound effect will play when a notification appears in the chat box for this item.",1,1,1,true)
    GameTooltip:Show()
end

function ConsumableTracker_SelectColorButton_OnEnter(self, motion)
    GameTooltip:SetOwner(self, "ANCHOR_TOPRIGHT")
    GameTooltip:SetText("Select Color")
    GameTooltip:AddLine("Select the color of notification's text. \nA preview of the text can be seen to the left.",1,1,1,true)
    GameTooltip:Show()
end

function ConsumableTracker_EnableRepeating_OnEnter(self, motion)
    GameTooltip:SetOwner(self, "ANCHOR_TOPRIGHT")
    GameTooltip:SetText("Set Repeating")
    GameTooltip:AddLine("If checked, the timer will repeat after sending a notification.",1,1,1,true)
    GameTooltip:Show()
end

function ConsumableTracker_EnableShowInRaids_OnEnter(self, motion)
    GameTooltip:SetOwner(self, "ANCHOR_TOPRIGHT")
    GameTooltip:SetText("Show In Raids")
    GameTooltip:AddLine("If checked, notifications will appear while in a raid group.",1,1,1,true)
    GameTooltip:Show()
end

function ConsumableTracker_EnableShowInGroups_OnEnter(self, motion)
    GameTooltip:SetOwner(self, "ANCHOR_TOPRIGHT")
    GameTooltip:SetText("Show In Groups")
    GameTooltip:AddLine("If checked, notifications will appear while in a group.",1,1,1,true)
    GameTooltip:Show()
end

function ConsumableTracker_EnableShowUngrouped_OnEnter(self, motion)
    GameTooltip:SetOwner(self, "ANCHOR_TOPRIGHT")
    GameTooltip:SetText("Show When Ungrouped")
    GameTooltip:AddLine("If checked, notifications will appear while not in a group.",1,1,1,true)
    GameTooltip:Show()
end


SLASH_CONSUMABLETRACKER1 = "/ct"
SLASH_CONSUMABLETRACKER2 = "/consumabletracker"
SlashCmdList["CONSUMABLETRACKER"] = ConsumableTracker_Commands